@@ -7,7 +7,7 @@ from pathlib import Path
 from typing import Iterable, Dict, Any, Sequence, Union
 from coba.backports import Literal
 
-from coba.pipes import IO, NullIO
+from coba.pipes import Sink, NullSink
 from coba.exceptions import CobaException
 from coba.registry import CobaRegistry
 from coba.utilities import coba_exit
@@ -15,8 +15,8 @@ from coba.utilities import coba_exit
 from coba.contexts.cachers import Cacher
 from coba.contexts.loggers import Logger
 
-# To support class properties before python 3.9 we must implement our properties directly 
-# on a meta class. Using class properties rather than class variables is done to allow 
+# To support class properties before python 3.9 we must implement our properties directly
+# on a meta class. Using class properties rather than class variables is done to allow
 # lazy loading. Lazy loading moves errors to execution time instead of import time where
 # they are easier to debug as well as removing import time circular references.
 
@@ -37,10 +37,11 @@ class ExperimentConfig:
     These can be overridden directly calling `config` on an Experiment.
     """
 
-    def __init__(self, processes:int, maxchunksperchild:int, chunk_by: Literal["source","task"]):
+    def __init__(self, processes:int, maxchunksperchild:int, maxtasksperchunk:int, chunk_by: Literal["source","task"]):
         """Instantiate an ExperimentConfig."""
         self.processes        : int                      = processes
         self.maxchunksperchild: int                      = maxchunksperchild
+        self.maxtasksperchunk : int                      = maxtasksperchunk
         self.chunk_by         : Literal["source","task"] = chunk_by
 
 class CobaContext_meta(type):
@@ -53,8 +54,8 @@ class CobaContext_meta(type):
     _cacher       = None
     _logger       = None
     _experiment   = None
-    _search_paths = [Path.home() , Path.cwd(), Path(sys.path[0]) ]
-    _store        = {}        
+    _search_paths = [Path.home() , Path.cwd(), Path(sys.path[0])]
+    _store        = {}
 
     def _load_file_configs(cls) -> Dict[str,Any]:
         config = {}
@@ -73,7 +74,7 @@ class CobaContext_meta(type):
                     cls._resolve_and_expand_paths(file_config, str(search_path))
 
                     config.update(file_config)
-                
+
                 except Exception as e:
                     raise CobaException(f"{str(e).strip('.')} in {potential_coba_config}.")
 
@@ -84,7 +85,7 @@ class CobaContext_meta(type):
             if isinstance(item, dict):
                 cls._resolve_and_expand_paths(item, current_dir)
 
-            if isinstance(item,str) and item.strip().startswith("~"):
+            if isinstance(item,str) and item.strip().startswith("~/"):
                 config_dict[key] = str(Path(item).expanduser().resolve())
 
             if isinstance(item,str) and (item.strip().startswith("../") or item.strip().startswith("./")):
@@ -101,12 +102,12 @@ class CobaContext_meta(type):
                     "api_keys"  : collections.defaultdict(lambda:None),
                     "cacher"    : { "DiskCacher": None},
                     "logger"    : { "IndentLogger": "Console" },
-                    "experiment": { "processes": 1, "maxchunksperchild": 0, "chunk_by": "source" }
+                    "experiment": { "processes": 1, "maxchunksperchild": 0, "maxtasksperchunk":0, "chunk_by": "source" }
                 }
 
                 for key,value in cls._load_file_configs().items():
                     if key in _raw_config and isinstance(_raw_config[key],dict) and not CobaRegistry.is_known_recipe(value):
-                       
+
                         if key == "experiment" and "maxtasksperchild" in value:
                             value["maxchunksperchild"] = value["maxtasksperchild"]
                             del value["maxtasksperchild"]
@@ -178,7 +179,7 @@ class CobaContext_meta(type):
         This store will be correctly marshalled to background processes if multiprocessing is used.
         """
         return cls._store
-    
+
     @store.setter
     def store(cls, value:Dict[str,Any]) -> None:
         cls._store = value
@@ -199,38 +200,34 @@ class CobaContext_meta(type):
         cls._search_paths = [ Path(path) if isinstance(path,str) else path for path in value  ]
 
 class CobaContext(metaclass=CobaContext_meta):
-    """To support class properties before python 3.9 we must implement our properties directly 
-       on a meta class. Using class properties rather than class variables is done to allow 
+    """To support class properties before python 3.9 we must implement our properties directly
+       on a meta class. Using class properties rather than class variables is done to allow
        lazy loading. Lazy loading moves errors to execution time instead of import time making
        them easier to debug as well as removing the potential for import time circular references.
     """ #I'm not sure some of the claims in the docstring are still true...
     pass
 
-class LearnerContext_meta(type):
+class InteractionContext_meta(type):
     """Global context scoped to the currently executing learner on a process.
 
-    At this time this exists primarily to allow Learners to log more in-depth information 
-    about their performance for algorithm debugging and analysis.
+    At this time this exists primarily to allow Learners to log more in-depth information about
+    their performance and state for algorithm debugging and analysis after an experiment is finished.
     """
-    _logger = None
+    _learner_info = dict()
 
     @property
-    def logger(cls) -> IO[Iterable[Dict[str,Any]],Dict[str,Any]]:
-        """Log information about a learner.
-        
-        These values are stored in the Result.interactions table. This means leaners
-        can log information on each learn/predict call and researchers can access it 
-        after the experiment for in-depth analysis.
+    def learner_info(cls) -> Dict[str,Any]:
+        """Information that can be accessed by EvaluationTasks to store information at Interaction granularity.
+
+        These values are stored in the Result.interactions table. This means leaners can log information on
+        each learn/predict call and researchers can access it after the experiment for in-depth analysis.
         """
-        return cls._logger if cls._logger is not None else NullIO[Dict[str,Any]]()
+        return cls._learner_info
 
-    @logger.setter
-    def logger(cls, value: IO[Iterable[Dict[str,Any]],Dict[str,Any]]):
-        cls._logger = value
-
-class LearnerContext(metaclass=LearnerContext_meta):
-    """To support class properties before python 3.9 we must implement our properties directly 
-       on a meta class. Using class properties rather than class variables is done to allow 
+class InteractionContext(metaclass=InteractionContext_meta):
+    """To support class properties before python 3.9 we must implement our properties directly
+       on a meta class. Using class properties rather than class variables is done to allow
        lazy loading. Lazy loading moves errors to execution time instead of import time making
        them easier to debug as well as removing the potential for import time circular references.
-    """ #I'm not sure some of the claims in the docstring are still true...
\ No newline at end of file
+    """ #I'm not sure some of the claims in the docstring are still true...
+    pass
