@@ -4,6 +4,7 @@ import json
 import time
 import collections.abc
 
+from numbers import Number
 from collections import Counter, OrderedDict, defaultdict
 from itertools import count, accumulate, chain
 from abc import ABC, abstractmethod
@@ -12,7 +13,7 @@ from typing import Iterator, Sequence, Generic, TypeVar, Any, Tuple, Union, Dict
 
 from coba.exceptions import CobaException
 
-_T_out = TypeVar('_T_out', bound=Any, covariant=True) 
+_T_out = TypeVar('_T_out', bound=Any, covariant=True)
 
 class Encoder(Generic[_T_out], ABC):
     """The Encoder interface.
@@ -186,7 +187,7 @@ class OneHotEncoder(Encoder[Tuple[int,...]]):
                 k[i] = 1
 
             keys_and_values = zip(values, map(tuple,known_onehots))
-    
+
             if self._err_if_unknown:
                 self._onehots = dict(keys_and_values)
             else:
@@ -210,7 +211,7 @@ class OneHotEncoder(Encoder[Tuple[int,...]]):
             raise CobaException(f"We were unable to find {e} in {list(self._onehots.keys())}")
 
     def encodes(self, values: Sequence[Any]) -> Sequence[Tuple[int,...]]:
-        
+
         if self._onehots is None:
             raise CobaException("This encoder must be fit before it can be used.")
 
@@ -259,7 +260,7 @@ class FactorEncoder(Encoder[int]):
             raise CobaException(f"We were unable to find {e} in {self._levels.keys()}") from None
 
     def encodes(self, values: Sequence[Any]) -> Sequence[int]:
-        
+
         if self._levels is None:
             raise CobaException("This encoder must be fit before it can be used.")
 
@@ -275,38 +276,47 @@ class CobaJsonDecoder(json.JSONDecoder):
     """A json decoder that allows for potential COBA extensions in the future."""
 
 class InteractionsEncoder:
-    
-    def __init__(self, interactions: Sequence[str]) -> None:
+
+    def __init__(self, interactions: Sequence[Union[str,float]]) -> None:
+        str_interactions = [i for i in interactions if isinstance(i,str)   ]
+        num_interactions = [i for i in interactions if isinstance(i,Number)]
 
         self.times       = [0,0,0,0]
         self.n           = 0
-        self._cross_pows = OrderedDict(zip(interactions,map(OrderedDict,map(Counter,interactions))))
-        self._ns_max_pow = { n:max(p.get(n,0) for p in self._cross_pows.values()) for n in set(''.join(interactions)) }
+        self._constant   = sum(num_interactions)
+        self._cross_pows = OrderedDict(zip(interactions,map(OrderedDict,map(Counter,str_interactions))))
+        self._ns_max_pow = { n:max(p.get(n,0) for p in self._cross_pows.values()) for n in set(''.join(str_interactions)) }
 
     def encode(self, **ns_raw_values: Union[str, float, Sequence[Union[str,float]], Dict[Union[str,int],Union[str,float]]]) -> Union[Sequence[float], Dict[str,float]]:
 
         self.n+= 1
 
-        is_sparse_type = lambda f: isinstance(f,dict) or isinstance(f,str)
-        is_sparse_sequ = lambda f: isinstance(f, collections.abc.Sequence) and any(map(is_sparse_type,f))
+        ns_raw_values = { k:v if v is not None else [] for k,v in ns_raw_values.items() }
+
+        is_str = lambda v: isinstance(v,str)
+        is_seq = lambda v: isinstance(v,collections.abc.Sequence) and not is_str(v)
+        is_map = lambda v: isinstance(v,collections.abc.Mapping)
+
+        is_sparse_type = lambda f: is_map(f) or is_str(f)
+        is_sparse_sequ = lambda f: is_seq(f) and any(map(is_sparse_type,f))
 
         is_sparse = any(is_sparse_type(v) or is_sparse_sequ(v) for v in ns_raw_values.values())
 
-        def make_all_dict_values(v) -> Dict[str,Union[str,float]]:
-            return v if isinstance(v,dict) else dict(zip(map(str,count()),v)) if isinstance(v,(list,tuple)) else { "0":v }
+        def make_dict(v) -> Dict[str,Union[str,float]]:
+            return v if is_map(v) else dict(zip(map(str,count()),v)) if is_seq(v) else { "0":v }
 
-        def make_all_list_values(v) -> Sequence[Union[str,float]]:
-            return v if isinstance(v, (list,tuple)) else [v]
+        def make_list(v) -> Sequence[Union[str,float]]:
+            return v if is_seq(v) else [v]
 
-        def handle_str_values(v: Dict[str,Union[str,float]]) -> Dict[str,float]:
-            return { (f"{x}{y}" if isinstance(y,str) else x):(1 if isinstance(y,str) else y) for x,y in v.items() }
+        def handle_str(v: Dict[str,Union[str,float]]) -> Dict[str,float]:
+            return { (f"{x}{y}" if is_str(y) else x):(1 if is_str(y) else y) for x,y in v.items() }
 
         start = time.time()
         if is_sparse:
-            ns_values = {ns:handle_str_values(make_all_dict_values(v)) for ns,v in ns_raw_values.items()}
-            ns_values = {ns:{f"{ns}{k}":v for k,v in values.items()}   for ns,values in ns_values.items() }
+            ns_values = { ns:handle_str(make_dict(V))            for ns,V in ns_raw_values.items() if ns in self._ns_max_pow }
+            ns_values = { ns:{f"{ns}{k}":v for k,v in V.items()} for ns,V in ns_values.items()     if ns in self._ns_max_pow }
         else:
-            ns_values = {ns:make_all_list_values(v) for ns,v in ns_raw_values.items()}
+            ns_values = { ns:make_list(v) for ns,v in ns_raw_values.items() if ns in self._ns_max_pow}
         self.times[0] += time.time()-start
 
         if is_sparse:
@@ -324,6 +334,8 @@ class InteractionsEncoder:
             encoded = dict(zip(chain.from_iterable(key_crosses), chain.from_iterable(val_crosses)))
             self.times[3] += time.time()-start
 
+            if self._constant: encoded['const'] = self._constant
+
             return encoded
         else:
             start = time.time()
@@ -338,11 +350,16 @@ class InteractionsEncoder:
             encoded = sum(val_crosses,[])
             self.times[3] += time.time()-start
 
+            if self._constant: encoded = [self._constant] + encoded
+
             return encoded
 
     def _pows(self, values, degree):
-        #WARNING: This function has been extremely optimized. Test speed before and after making any changes.
-        #WARNING: Look in test_performance for three existing performance tests.
+        #WARNING: This function has been extremely optimized. Please baseline performance before and after making any changes.
+        #WARNING: You can find three existing performance tests in test_performance.
+
+        if values in [[],{}]: return []
+
         starts = [1]*len(values)
         terms  = [['']] if isinstance(values[0],str) else [[1]]
 
@@ -360,6 +377,8 @@ class InteractionsEncoder:
         #WARNING: This function has been extremely optimized. Test speed before and after making any changes.
         #WARNING: Look in test_performance for three existing performance tests.
 
+        if any(ns_pows[k] == [] for k in cross_pow.keys()): return []
+
         values = [ ns_pows[ns][p] for ns,p in cross_pow.items() ]
         cross  = values[0]
 
