@@ -137,9 +137,6 @@ class VowpalMediator:
         self._curr_ns_offset += length
         return value
 
-    def __reduce__(self):
-        return (VowpalMediator, ())
-
 class VowpalArgsLearner(Learner):
     """A friendly wrapper around Vowpal Wabbit's python interface to support CB learning.
 
